#!/usr/bin/env python3
import logging
import os
from pathlib import Path
from typing import Tuple, Optional

import click

from clin import __version__
from clin.clients.nakadi_sql import NakadiSql
from clin.clinfile import calculate_scope
from clin.config import ConfigurationError, load_config
from clin.clients.nakadi import Nakadi, NakadiError
from clin.models.shared import Kind
from clin.processor import Processor, ProcessingError
from clin.utils import configure_logging, pretty_yaml, pretty_json
from clin.yamlops import YamlLoader, load_manifest, load_yaml, YamlError

DEFAULT_YAML_LOADER = YamlLoader()


@click.group()
@click.version_option(version=__version__)
def cli():
    """Manage Nakadi resources"""
    pass


@cli.command("apply")
@click.option(
    "-t",
    "--token",
    required=False,
    type=str,
    help="The bearer token to authenticate the Nakadi requests",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose output (default - false)",
)
@click.option(
    "-e",
    "--env",
    required=True,
    type=str,
    help="The Nakadi environment to target",
)
@click.option(
    "-X",
    "--execute",
    is_flag=True,
    default=False,
    help="Execute updates (default - false)",
)
@click.option(
    "-d",
    "--show-diff",
    is_flag=True,
    default=False,
    help="Display the schema diff (default - false)",
)
@click.option(
    "-p",
    "--show-payload",
    is_flag=True,
    default=False,
    help="Show Nakadi payload (default - false)",
)
@click.argument("file", type=click.Path(exists=True, dir_okay=False, readable=True))
def apply(
    token: Optional[str],
    verbose: bool,
    env: str,
    execute: bool,
    show_diff: bool,
    show_payload: bool,
    file: str,
):
    """Create or update Nakadi resource from single yaml manifest file\n
    Values to fill {{VARIABLES}} are taken from system environment"""
    configure_logging(verbose)

    try:
        config = load_config()
        envelope = load_manifest(Path(file), DEFAULT_YAML_LOADER, os.environ)
        processor = Processor(config, token, execute, show_diff, show_payload)
        processor.apply(env, envelope)

    except (ProcessingError, NakadiError, ConfigurationError, YamlError) as ex:
        logging.error(ex)
        exit(-1)

    except Exception as ex:
        logging.exception(ex)
        exit(-1)


@cli.command("process")
@click.option(
    "-t",
    "--token",
    required=False,
    type=str,
    help="The bearer token to authenticate the Nakadi requests",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose output (default - false)",
)
@click.option(
    "-X",
    "--execute",
    is_flag=True,
    default=False,
    help="Execute the updates (default - false)",
)
@click.option(
    "-d",
    "--show-diff",
    is_flag=True,
    default=False,
    help="Display the schema diff (default - false)",
)
@click.option(
    "-p",
    "--show-payload",
    is_flag=True,
    default=False,
    help="Show Nakadi payload (default - false)",
)
@click.option(
    "-i",
    "--id",
    required=False,
    type=str,
    multiple=True,
    help="Select one or multiple steps to process by their id",
)
@click.option(
    "-e",
    "--env",
    required=False,
    type=str,
    multiple=True,
    help="Select one or multiple steps to process by matching the target environment",
)
@click.argument("file", type=click.Path(exists=True, dir_okay=False, readable=True))
def process(
    token: Optional[str],
    verbose: bool,
    execute: bool,
    show_diff: bool,
    show_payload: bool,
    id: Tuple[str],
    env: Tuple[str],
    file: str,
):
    """Create or update multiple Nakadi resources from a clin file"""
    configure_logging(verbose)

    try:
        config = load_config()
        processor = Processor(config, token, execute, show_diff, show_payload)
        file_path: Path = Path(file)
        master = load_yaml(file_path, DEFAULT_YAML_LOADER, os.environ)

        scope = calculate_scope(master, file_path.parent, DEFAULT_YAML_LOADER, id, env)

        for task in (
            scope[Kind.EVENT_TYPE] + scope[Kind.SQL_QUERY] + scope[Kind.SUBSCRIPTION]
        ):
            logging.debug(
                "[%s] applying file %s to %s environment",
                task.id,
                task.path,
                task.target,
            )
            processor.apply(task.target, task.envelope)

    except (ProcessingError, ConfigurationError, YamlError) as ex:
        logging.error(ex)
        exit(-1)

    except Exception as ex:
        logging.exception(ex)
        exit(-1)


@cli.command("dump")
@click.option(
    "-t",
    "--token",
    required=False,
    type=str,
    help="The bearer token to authenticate the Nakadi requests",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    help="Verbose output (default - false)",
)
@click.option(
    "--include-envelope",
    is_flag=True,
    default=False,
    help="Add clin envelope to schema (default - false)",
)
@click.option(
    "-e",
    "--env",
    required=True,
    type=str,
    help="The Nakadi environment to target",
)
@click.option(
    "-o",
    "--output",
    default="yaml",
    type=click.Choice(["yaml", "json"]),
    help="The output format (default - yaml)",
)
@click.argument("event_type", type=str)
def dump(
    token: Optional[str],
    verbose: bool,
    env: str,
    output: str,
    include_envelope: bool,
    event_type: str,
):
    """Print manifest of existing Nakadi event type"""
    configure_logging(verbose)

    try:
        config = load_config()
        if env not in config.environments:
            logging.error(f"Environment not found in configuration: {env}")
            exit(-1)

        nakadi = Nakadi(config.environments[env].nakadi_url, token)
        entity = nakadi.get_event_type(event_type)

        if entity and config.environments[env].nakadi_sql_url:
            nakadi_sql = NakadiSql(config.environments[env].nakadi_sql_url, token)
            entity = nakadi_sql.get_sql_query(entity) or entity

        if entity is None:
            logging.error("Event type not found in Nakadi %s: %s", env, event_type)
            exit(-1)

        payload = (
            entity.to_envelope().to_manifest() if include_envelope else entity.to_spec()
        )

        if output.lower() == "yaml":
            logging.info(pretty_yaml(payload))
        elif output.lower() == "json":
            logging.info(pretty_json(payload))
        else:
            logging.error("Invalid output format: %s", output)
            exit(-1)

    except (NakadiError, ConfigurationError) as ex:
        logging.error(ex)
        exit(-1)

    except Exception as ex:
        logging.exception(ex)
        exit(-1)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import json
from typing import List, Optional

import requests
from requests import HTTPError, Response

from clin.clients.http_client import (
    HttpClient,
    rw_auth_from_payload,
    ro_auth_from_payload,
    auth_to_payload,
)
from clin.models.event_type import (
    EventType,
    Category,
    Cleanup,
    Partitioning,
    Schema,
    Audience,
)
from clin.models.subscription import Subscription
from clin.utils import MS_IN_DAY


class Nakadi(HttpClient):
    def get_event_type(self, name: str) -> Optional[EventType]:
        try:
            payload = self._get(f"event-types/{name}")
            partition_count = self.get_partition_count(name)
            return event_type_from_payload(payload, partition_count)

        except HTTPError as e:
            if e.response.status_code == 404:
                return None
            raise NakadiError(
                f"Nakadi error during getting event type '{name}'", e.response
            )

    def get_partition_count(self, name: str) -> int:
        try:
            return len(self._get(f"event-types/{name}/partitions"))

        except HTTPError as e:
            raise NakadiError(
                f"Can not get partitions for event type'{name}'", e.response
            )

    def create_event_type(self, event_type: EventType):
        resp = self._post(
            "event-types", data=json.dumps(event_type_to_payload(event_type))
        )
        if resp.status_code != 201:
            raise NakadiError(
                f"Nakadi error during creation of event type '{event_type.name}'", resp
            )

    def update_event_type(self, event_type: EventType):
        resp = self._put(
            f"event-types/{event_type.name}",
            data=json.dumps(event_type_to_payload(event_type)),
        )
        if resp.status_code != 200:
            raise NakadiError(
                f"Nakadi error during updating of event type '{event_type.name}'", resp
            )

    def get_subscription(
        self, event_types: List, owning_application: str, consumer_group: str
    ) -> Optional[Subscription]:
        params_str = Subscription.components_string(
            event_types, owning_application, consumer_group
        )

        url = f"{self._base_url}/subscriptions"
        params = {
            "event_type": ",".join(event_types),
            "owning_application": owning_application,
        }
        resp = requests.get(url, headers=self._headers, params=params)
        if resp.status_code != 200:
            raise NakadiError(
                f"Nakadi error during getting subscription for {params_str}", resp
            )

        subscriptions = [
            s for s in resp.json()["items"] if s["consumer_group"] == consumer_group
        ]
        if len(subscriptions) > 1:
            raise NakadiError(f"Got multiple subscriptions for {params_str}", resp)
        elif len(subscriptions) == 1:
            payload = subscriptions[0]
            sub = subscription_from_payload(payload)
            return sub
        else:
            return None

    def create_subscription(self, subscription: Subscription) -> Subscription:
        payload = json.dumps(subscription_to_payload(subscription))
        resp = requests.post(
            f"{self._base_url}/subscriptions", headers=self._headers, data=payload
        )
        if resp.status_code != 201:
            raise NakadiError(f"Nakadi error during creating {subscription}", resp)

        created = subscription_from_payload(resp.json())
        return created

    def update_subscription(self, subscription: Subscription):
        payload = json.dumps(subscription_to_payload(subscription))
        resp = requests.put(
            f"{self._base_url}/subscriptions/{subscription.id}",
            headers=self._headers,
            data=payload,
        )
        if resp.status_code != 204:
            raise NakadiError(f"Nakadi error during updating {subscription}", resp)


class NakadiError(Exception):
    def __init__(self, message: str, response: Response):
        self.response = response
        self._message = message

    def __str__(self):
        code = self.response.status_code
        msg = f"{self._message}: {code}"

        try:
            body = self.response.json()
            if "problem" in body:
                return msg + " - " + body["problem"]["detail"]
            else:
                return msg + " - " + body

        except Exception:
            if self.response.text:
                return msg + " - " + self.response.text
            else:
                return msg


def event_type_to_payload(event_type: EventType) -> dict:
    enrichment_strategies = (
        [] if event_type.category == Category.UNDEFINED else ["metadata_enrichment"]
    )
    return {
        "name": event_type.name,
        "owning_application": event_type.owning_application,
        "category": str(event_type.category),
        "audience": str(event_type.audience),
        "partition_strategy": str(event_type.partitioning.strategy),
        "partition_key_fields": event_type.partitioning.keys,
        "default_statistic": {
            "messages_per_minute": 100,
            "message_size": 100,
            "read_parallelism": event_type.partitioning.partition_count,
            "write_parallelism": event_type.partitioning.partition_count,
        },
        "cleanup_policy": event_type.cleanup.policy,
        "options": {
            "retention_time": event_type.cleanup.retention_time_days * MS_IN_DAY
        },
        "compatibility_mode": event_type.schema.compatibility,
        "schema": {
            "type": "json_schema",
            "version": "1.0.0",
            "schema": json.dumps(event_type.schema.json_schema),
        },
        "authorization": auth_to_payload(event_type.auth),
        "enrichment_strategies": enrichment_strategies,
    }


def event_type_from_payload(payload: dict, partition_count: int) -> EventType:
    return EventType(
        name=payload["name"],
        category=Category(payload["category"]),
        owning_application=payload["owning_application"],
        audience=Audience(payload["audience"]) if "audience" in payload else None,
        partitioning=Partitioning(
            strategy=Partitioning.Strategy(payload["partition_strategy"]),
            keys=payload.get("partition_key_fields"),
            partition_count=partition_count,
        ),
        cleanup=Cleanup(
            policy=Cleanup.Policy(payload["cleanup_policy"]),
            retention_time_days=payload["options"].get("retention_time", 0)
            // MS_IN_DAY,
        ),
        schema=Schema(
            compatibility=Schema.Compatibility(payload["compatibility_mode"]),
            json_schema=json.loads(payload["schema"]["schema"]),
        ),
        auth=rw_auth_from_payload(payload["authorization"]),
    )


def subscription_to_payload(subscription: Subscription) -> dict:
    return {
        "owning_application": subscription.owning_application,
        "event_types": subscription.event_types,
        "consumer_group": subscription.consumer_group,
        "initial_cursors": [],
        "read_from": "end",
        "authorization": auth_to_payload(subscription.auth),
    }


def subscription_from_payload(payload: dict) -> Subscription:
    return Subscription(
        id=payload["id"],
        owning_application=payload["owning_application"],
        event_types=payload["event_types"],
        consumer_group=payload["consumer_group"],
        auth=ro_auth_from_payload(payload["authorization"]),
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass

from clin.utils import ensure_flat_list


@dataclass
class Auth(ABC):
    users: dict[str, list[str]]
    services: dict[str, list[str]]
    any_token: dict[str, bool]

    @classmethod
    @abstractmethod
    def get_roles(cls) -> list[str]:
        pass

    @classmethod
    @abstractmethod
    def get_any_token_values(cls) -> list[str]:
        pass

    @classmethod
    def from_spec(cls, spec: dict) -> Auth:
        return cls(
            users=cls._parse_section(spec, "users"),
            services=cls._parse_section(spec, "services"),
            any_token=cls._parse_any_token(spec),
        )

    def to_spec(self) -> dict[str, dict[str, list[str]]]:
        return {
            "users": {role: self.users[role] for role in self.get_roles()},
            "services": {role: self.services[role] for role in self.get_roles()},
            "anyToken": {
                role: self.any_token.get(role, False)
                for role in self.get_any_token_values()
            },
        }

    @classmethod
    def _parse_section(cls, spec: dict, section: str) -> dict[str, list[str]]:
        def parse(role: str):
            return (
                list(set(ensure_flat_list(spec[section].get(role))))
                if section in spec and spec[section]
                else []
            )

        return {role: parse(role) for role in cls.get_roles()}

    @classmethod
    def _parse_any_token(cls, spec: dict) -> dict[str, bool]:
        return {
            role: spec.get("anyToken", {}).get(role, False)
            for role in cls.get_any_token_values()
        }


@dataclass
class ReadOnlyAuth(Auth):
    @classmethod
    def get_roles(cls) -> list[str]:
        return ["admins", "readers"]

    @classmethod
    def get_any_token_values(cls) -> list[str]:
        return ["read"]


@dataclass
class ReadWriteAuth(Auth):
    @classmethod
    def get_roles(cls) -> list[str]:
        return ["admins", "readers", "writers"]

    @classmethod
    def get_any_token_values(cls) -> list[str]:
        return ["read", "write"]



from __future__ import annotations

from dataclasses import dataclass

from colorama import Fore

from clin.models.auth import ReadWriteAuth
from clin.models.shared import (
    Cleanup,
    Category,
    Entity,
    Kind,
    Audience,
    Partitioning,
    Schema,
)


@dataclass
class EventType(Entity):
    name: str
    category: Category
    owning_application: str
    audience: Audience
    partitioning: Partitioning
    cleanup: Cleanup
    schema: Schema
    auth: ReadWriteAuth

    def __str__(self) -> str:
        return f"event type {Fore.BLUE}{self.name}{Fore.RESET}"

    @property
    def kind(self) -> Kind:
        return Kind.EVENT_TYPE

    @staticmethod
    def from_spec(spec: dict[str, any]) -> EventType:
        return EventType(
            name=spec["name"],
            category=Category(spec["category"]),
            owning_application=spec["owningApplication"],
            audience=Audience(spec["audience"]),
            partitioning=Partitioning.from_spec(spec["partitioning"]),
            cleanup=Cleanup.from_spec(spec["cleanup"]),
            schema=Schema.from_spec(spec["schema"]),
            auth=ReadWriteAuth.from_spec(spec["auth"]),
        )

    def to_spec(self) -> dict[str, any]:
        return {
            "name": self.name,
            "category": str(self.category),
            "owningApplication": self.owning_application,
            "audience": str(self.audience),
            "partitioning": self.partitioning.to_spec(),
            "cleanup": self.cleanup.to_spec(),
            "schema": self.schema.to_spec(),
            "auth": self.auth.to_spec() if self.auth else {},
        }



import logging
import os
from typing import Optional, TypeVar

import requests
from requests import Response

from clin import __version__
from clin.models.auth import Auth, ReadOnlyAuth, ReadWriteAuth

TAuth = TypeVar("TAuth", bound=Auth)


class HttpClient:
    def __init__(self, base_url: str, token: Optional[str]):
        self._base_url = base_url.rstrip("/")
        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": f"clin+{__version__}",
        }
        if token:
            self._headers["Authorization"] = f"Bearer {token}"

    def _get(self, path: str, **kwargs) -> dict:
        url = os.path.join(self._base_url, path)
        logging.debug(f"GET {url}")
        resp = requests.get(url, headers=self._headers, **kwargs)
        logging.debug(f"-> response code {resp.status_code}")
        resp.raise_for_status()
        return resp.json()

    def _post(self, path: str, **kwargs) -> Response:
        url = os.path.join(self._base_url, path)
        logging.debug(f"POST {url}")
        resp = requests.post(url, headers=self._headers, **kwargs)
        logging.debug(f"-> response code {resp.status_code}")
        return resp

    def _put(self, path: str, **kwargs) -> Response:
        url = os.path.join(self._base_url, path)
        logging.debug(f"PUT {url}")
        resp = requests.put(url, headers=self._headers, **kwargs)
        logging.debug(f"-> response code {resp.status_code}")
        return resp


def auth_to_payload(auth: Auth) -> dict:
    def parse(role: str):
        def el(key: str):
            return [
                {"data_type": key, "value": x} for x in getattr(auth, key + "s")[role]
            ]

        return el("user") + el("service")

    payload = {role: parse(role) for role in auth.get_roles()}
    if auth.any_token.get("read", False):
        payload["readers"].append({"data_type": "*", "value": "*"})
    if auth.any_token.get("write", False):
        payload["writers"].append({"data_type": "*", "value": "*"})

    return payload


def ro_auth_from_payload(payload: dict) -> Optional[ReadOnlyAuth]:
    return _auth_from_payload(ReadOnlyAuth({}, {}, {}), payload)


def rw_auth_from_payload(payload: dict) -> Optional[ReadWriteAuth]:
    return _auth_from_payload(ReadWriteAuth({}, {}, {}), payload)


def _auth_from_payload(auth: TAuth, payload: dict) -> Optional[TAuth]:
    if not payload:
        return None

    for role in auth.get_roles():
        auth.users[role] = []
        auth.services[role] = []
        for el in payload.get(role, []):
            if el["data_type"] == "user":
                auth.users[role].append(el["value"])
            if el["data_type"] == "service":
                auth.services[role].append(el["value"])
            if el["data_type"] == "*":
                if role == "readers" and "read" in auth.get_any_token_values():
                    auth.any_token["read"] = True
                if role == "writers" and "write" in auth.get_any_token_values():
                    auth.any_token["write"] = True

    return auth



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import unique, Enum
from typing import Optional


@unique
class Audience(str, Enum):
    COMPONENT_INTERNAL = "component-internal"
    BUSINESS_UNIT_INTERNAL = "business-unit-internal"
    COMPANY_INTERNAL = "company-internal"
    EXTERNAL_PARTNER = "external-partner"
    EXTERNAL_PUBLIC = "external-public"

    def __str__(self) -> str:
        return str(self.value)


@unique
class Category(str, Enum):
    BUSINESS = "business"
    DATA = "data"
    UNDEFINED = "undefined"

    def __str__(self) -> str:
        return str(self.value)


@unique
class Kind(str, Enum):
    EVENT_TYPE = "event-type"
    SQL_QUERY = "sql-query"
    SUBSCRIPTION = "subscription"

    def __str__(self) -> str:
        return str(self.value)


@dataclass
class Cleanup:
    @unique
    class Policy(str, Enum):
        DELETE = "delete"
        COMPACT = "compact"

        def __str__(self) -> str:
            return str(self.value)

    policy: Policy
    retention_time_days: int

    @staticmethod
    def from_spec(spec: dict[str, any]) -> Cleanup:
        return Cleanup(
            policy=Cleanup.Policy(spec["policy"]),
            retention_time_days=spec.get(
                "retentionTimeDays", 1  # 1 is default Nakadi value
            ),
        )

    def to_spec(self) -> dict[str, any]:
        return {
            "policy": str(self.policy),
            "retentionTimeDays": self.retention_time_days,
        }


@dataclass
class Partitioning:
    @unique
    class Strategy(str, Enum):
        USER_DEFINED = "user_defined"
        HASH = "hash"
        RANDOM = "random"

        def __str__(self) -> str:
            return str(self.value)

    strategy: Strategy
    keys: Optional[list[str]]
    partition_count: int

    @staticmethod
    def from_spec(spec: dict[str, any]) -> Partitioning:
        return Partitioning(
            strategy=Partitioning.Strategy(spec["strategy"]),
            keys=spec.get("keys"),
            partition_count=spec["partitionCount"],
        )

    def to_spec(self) -> dict[str, any]:
        return {
            "strategy": str(self.strategy),
            "keys": self.keys,
            "partitionCount": self.partition_count,
        }


@dataclass
class Schema:
    @unique
    class Compatibility(str, Enum):
        NONE = "none"
        FORWARD = "forward"
        COMPATIBLE = "compatible"

        def __str__(self) -> str:
            return str(self.value)

    compatibility: Compatibility
    json_schema: dict[str, any]

    @staticmethod
    def from_spec(spec: dict[str, any]) -> Schema:
        return Schema(
            compatibility=Schema.Compatibility(spec["compatibility"]),
            json_schema=spec["jsonSchema"],
        )

    def to_spec(self) -> dict[str, any]:
        return {
            "compatibility": str(self.compatibility),
            "jsonSchema": self.json_schema,
        }


@dataclass
class Envelope:
    kind: Kind
    spec: dict[str, any]

    @staticmethod
    def from_manifest(manifest: dict[str, any]) -> Envelope:
        if "kind" not in manifest:
            raise ValueError("Required field `kind` not found")
        if "spec" not in manifest:
            raise ValueError(f"Required field `spec` not found")

        return Envelope(kind=Kind(manifest["kind"]), spec=manifest["spec"])

    def to_manifest(self) -> dict[str, any]:
        return {
            "kind": str(self.kind),
            "spec": self.spec,
        }


class Entity(ABC):
    @property
    @abstractmethod
    def kind(self) -> Kind:
        pass

    @abstractmethod
    def to_spec(self) -> dict[str, any]:
        pass

    def to_envelope(self) -> Envelope:
        return Envelope(kind=self.kind, spec=self.to_spec())



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from colorama import Fore

from clin.models.auth import ReadOnlyAuth
from clin.models.shared import Cleanup, Category, Entity, Kind, Audience, Partitioning


@dataclass
class OutputEventType:
    category: Category
    owning_application: str
    audience: Audience
    repartitioning: Optional[Partitioning]
    cleanup: Cleanup

    @staticmethod
    def from_spec(spec: dict[str, any]):
        return OutputEventType(
            category=Category(spec["category"]),
            owning_application=spec["owningApplication"],
            audience=Audience(spec["audience"]),
            repartitioning=Partitioning.from_spec(spec["repartitioning"])
            if "repartitioning" in spec
            else None,
            cleanup=Cleanup.from_spec(spec["cleanup"]),
        )

    def to_spec(self) -> dict[str, any]:
        spec = {
            "category": str(self.category),
            "owningApplication": self.owning_application,
            "audience": str(self.audience),
            "cleanup": self.cleanup.to_spec(),
        }

        if self.repartitioning:
            spec["repartitioning"] = self.repartitioning.to_spec()

        return spec


@dataclass
class SqlQuery(Entity):
    name: str
    sql: str
    envelope: bool
    output_event_type: OutputEventType
    auth: ReadOnlyAuth

    def __str__(self) -> str:
        return f"sql query {Fore.BLUE}{self.name}{Fore.RESET}"

    @property
    def kind(self) -> Kind:
        return Kind.SQL_QUERY

    @staticmethod
    def from_spec(spec: dict[str, any]) -> SqlQuery:
        return SqlQuery(
            name=spec["name"],
            sql=spec["sql"],
            envelope=spec["envelope"],
            output_event_type=OutputEventType.from_spec(spec["outputEventType"]),
            auth=ReadOnlyAuth.from_spec(spec["auth"]),
        )

    def to_spec(self) -> dict[str, any]:
        return {
            "name": self.name,
            "sql": self.sql,
            "envelope": self.envelope,
            "outputEventType": self.output_event_type.to_spec(),
            "auth": self.auth.to_spec() if self.auth else {},
        }



import json
from typing import Optional

from requests import HTTPError

from clin.clients.nakadi import NakadiError
from clin.clients.http_client import HttpClient, ro_auth_from_payload, auth_to_payload
from clin.models.event_type import EventType
from clin.models.sql_query import SqlQuery, OutputEventType
from clin.models.shared import Category, Cleanup, Audience, Partitioning
from clin.utils import MS_IN_DAY


class NakadiSql(HttpClient):
    def get_sql_query(self, event_type: EventType) -> Optional[SqlQuery]:
        try:
            payload = self._get(f"queries/{event_type.name}")
            return sql_query_from_payload(event_type, payload)

        except HTTPError as e:
            if e.response.status_code == 404:
                return None
            raise NakadiError(
                f"Nakadi error trying to get sql query '{event_type}'", e.response
            )

    def create_sql_query(self, sql_query: SqlQuery):
        resp = self._post("queries", data=json.dumps(sql_query_to_payload(sql_query)))
        if resp.status_code != 201:
            raise NakadiError(
                f"Nakadi error during creation of sql query '{sql_query.name}'", resp
            )

    def update_sql_query_auth(self, sql_query: SqlQuery):
        resp = self._put(
            f"queries/{sql_query.name}/authorization",
            data=json.dumps(auth_to_payload(sql_query.auth)),
        )
        if resp.status_code != 200:
            raise NakadiError(
                f"Nakadi error during updating of sql query '{sql_query.name}'", resp
            )


def output_event_type_from_payload(
    event_type: EventType, payload: dict
) -> OutputEventType:
    def convert_repartitioning() -> Optional[Partitioning]:
        return (
            Partitioning(
                strategy=Partitioning.Strategy(
                    payload["repartition_parameters"]["partition_strategy"]
                ),
                keys=payload["repartition_parameters"].get("partition_key_fields"),
                partition_count=int(
                    payload["repartition_parameters"]["number_of_partitions"]
                ),
            )
            if "repartition_parameters" in payload
            else None
        )

    return OutputEventType(
        category=Category(payload["category"]),
        owning_application=payload.get(
            "owning_application", event_type.owning_application
        ),
        audience=Audience(payload["audience"])
        if "audience" in payload
        else event_type.audience,
        repartitioning=convert_repartitioning(),
        cleanup=Cleanup(
            policy=Cleanup.Policy(payload["cleanup_policy"]),
            retention_time_days=payload.get("retention_time", 0) // MS_IN_DAY,
        ),
    )


def sql_query_from_payload(event_type: EventType, payload: dict) -> SqlQuery:
    if payload["id"] != payload["output_event_type"]["name"]:
        raise NakadiError(
            "The output event type's name does not match the sql query id."
            " This is unexpected and unsupported - please report this issue in the clin project on GitHub!"
        )

    return SqlQuery(
        name=payload["id"],
        sql=payload["sql"],
        envelope=payload["envelope"],
        output_event_type=output_event_type_from_payload(
            event_type, payload["output_event_type"]
        ),
        auth=ro_auth_from_payload(payload["authorization"]),
    )


def sql_query_to_payload(sql_query: SqlQuery) -> dict:
    payload = {
        "id": sql_query.name,
        "sql": sql_query.sql,
        "envelope": sql_query.envelope,
        "output_event_type": {
            "name": sql_query.name,
            "owning_application": sql_query.output_event_type.owning_application,
            "category": str(sql_query.output_event_type.category),
            "audience": str(sql_query.output_event_type.audience),
            "cleanup_policy": str(sql_query.output_event_type.cleanup.policy),
            "retention_time": sql_query.output_event_type.cleanup.retention_time_days
            * MS_IN_DAY,
        },
        "authorization": auth_to_payload(sql_query.auth),
    }

    if sql_query.output_event_type.repartitioning:
        payload["repartition_parameters"] = {
            "number_of_partitions": sql_query.output_event_type.repartitioning.partition_count,
            "partition_strategy": str(
                sql_query.output_event_type.repartitioning.strategy
            ),
            "partition_key_fields": sql_query.output_event_type.repartitioning.keys,
        }

    return payload



from __future__ import annotations

import logging
from dataclasses import dataclass
from itertools import chain
from pathlib import Path

from clin.models.shared import Envelope, Kind
from clin.yamlops import YamlLoader


@dataclass
class Process:
    id: str
    path: str
    envelope: Envelope
    target: str


def calculate_scope(
    master: dict,
    base_path: Path,
    loader: YamlLoader,
    filter_id: tuple[str],
    filter_env: tuple[str],
) -> dict[Kind, list[Process]]:
    processes = master.get("process")
    if not processes:
        raise Exception("'process' section is not found")

    scope = {kind: [] for kind in Kind}
    for proc in processes:
        if len(filter_id) > 0 and proc["id"] not in filter_id:
            continue
        if len(filter_env) > 0 and proc["target"] not in filter_env:
            continue

        proc_id = proc["id"]
        sources = [base_path.joinpath(s).resolve(strict=False) for s in proc["paths"]]
        manifests_files = []
        for source in sources:
            if not source.exists():
                raise Exception(
                    f"Specified path 'f{source.absolute()}' is not found for process '{proc_id}'"
                )

            manifests_files += chain(source.glob("*.yml"), source.glob("*.yaml"))
            if not manifests_files:
                logging.warning(
                    "No manifests found for process '%s' in '%s'",
                    proc_id,
                    source.absolute(),
                )

        for manifest_file in manifests_files:
            manifest = loader.load_yaml_from_file(manifest_file, proc.get("env", {}))
            envelope = Envelope.from_manifest(manifest)
            scope[envelope.kind].append(
                Process(
                    id=proc_id,
                    path=manifest_file,
                    envelope=envelope,
                    target=proc["target"],
                )
            )

    return scope



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['clin', 'clin.clients', 'clin.models']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1,<8.0',
 'colorama>=0.4,<0.5',
 'deepdiff>=5.2,<6.0',
 'pygments>=2.6,<3.0',
 'pyyaml>=5.4,<6.0',
 'requests>=2.23,<3.0']

entry_points = \
{'console_scripts': ['clin = clin.run:cli']}

setup_kwargs = {
    'name': 'clin',
    'version': '1.2.0',
    'description': 'Cli for Nakadi resources management in Infrastructure as Code style',
    'long_description': '# Clin\n\n**Clin** is a command-line utility to manage [Nakadi](https://nakadi.io/)\nresources from schema files in "Infrastructure as Code" style.\n![](https://github.com/zalando-incubator/clin/raw/master/docs/gifs/demo.gif)\n\n## User Guide\n\n### Prerequisites\n\n* [Python](https://www.python.org/) >= 3.7\n\n### Installing\nYou can install **clin** directly from [PyPI](https://pypi.org/project/clin/)\nusing pip:\n\n```bash\npip install clin\n```\n\n### Getting started\n\nAfter this you should be able to run the `clin` tool:\n```bash\n~ clin --help\nUsage: clin [OPTIONS] COMMAND [ARGS]...\n...\n```\n\nPlease refer to the [documentation](https://github.com/zalando-incubator/clin/tree/master/docs)\nand [examples](https://github.com/zalando-incubator/clin/tree/master/docs/examples).\n\n## Contributing\n\nPlease read [CONTRIBUTING](https://github.com/zalando-incubator/clin/blob/master/CONTRIBUTING.md)\nfor details on our process for submitting pull requests to us, and please ensure\nyou follow the [CODE_OF_CONDUCT](https://github.com/zalando-incubator/clin/blob/master/CODE_OF_CONDUCT.md).\n\n### Prerequisites\n\n* [Python](https://www.python.org/) >= 3.7\n* [Poetry](https://python-poetry.org/) for packaging and dependency\n  management. See the [official docs](https://python-poetry.org/docs/) for\n  instructions on installation and basic usage.\n\n### Installing\nAfter cloning the repository, use `poetry` to create a new virtual environment\nand restore all dependencies.\n\n```bash\npoetry install\n```\n\nIf you\'re using an IDE (eg. PyCharm), make sure that it\'s configured to use the\nvirtual environment created by poetry as the project\'s interpreter. You can find\nthe path to the used environment with `poetry env info`.\n\n### Running the tests\n\n```bash\npoetry run pytest\n```\n\n## Versioning\n\nWe use [SemVer](http://semver.org) for versioning. For the versions available,\nsee the [tags on this repository](https://github.com/zalando-incubator/clin/tags).\n\n## Authors\n\n* **Dmitry Erokhin** [@Dmitry-Erokhin](https://github.com/Dmitry-Erokhin)\n* **Daniel Stockhammer** [@dstockhammer](https://github.com/dstockhammer)\n\nSee also the list of [contributors](Chttps://github.com/zalando-incubator/clin/blob/master/ONTRIBUTORS.md)\nwho participated in this project.\n\n## License\n\nThis project is licensed under the MIT License. See the\n[LICENSE](https://github.com/zalando-incubator/clin/blob/master/LICENSE) file\nfor details.\n',
    'author': 'Dmitry Erokhin',
    'author_email': 'dmitry.erokhin@zalando.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zalando-incubator/clin',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
import logging
from typing import Optional, Dict, Callable

from colorama import Fore
from deepdiff import DeepDiff

from clin.clients.nakadi import (
    Nakadi,
    NakadiError,
    event_type_to_payload,
    subscription_to_payload,
)
from clin.clients.nakadi_sql import NakadiSql, sql_query_to_payload
from clin.config import AppConfig
from clin.models.auth import ReadWriteAuth, ReadOnlyAuth
from clin.models.event_type import EventType
from clin.models.shared import Kind, Envelope, Entity
from clin.models.sql_query import SqlQuery
from clin.models.subscription import Subscription
from clin.utils import pretty_yaml, pretty_json

MODIFY_COLOR = Fore.MAGENTA
ERROR_COLOR = Fore.RED
UP_TO_DATE_COLOR = Fore.GREEN
OUTPUT_INDENTATION = 4


class Processor:
    def __init__(
        self,
        config: AppConfig,
        token: Optional[str],
        execute: bool = False,
        show_diff: bool = False,
        show_payload: bool = False,
    ):
        self.apply_func_per_kind: Dict[Kind, Callable[[str, dict], None]] = {
            Kind.EVENT_TYPE: self.apply_event_type,
            Kind.SQL_QUERY: self.apply_sql_query,
            Kind.SUBSCRIPTION: self.apply_subscription,
        }

        self.token = token
        self.config = config
        self.execute = execute
        self.show_diff = show_diff
        self.show_payload = show_payload

    def apply(self, env: str, envelope: Envelope):
        apply = self.apply_func_per_kind.get(envelope.kind, None)
        if apply is None:
            raise ProcessingError(f"Unsupported kind: {envelope.kind}")
        apply(env, envelope.spec)

    def apply_event_type(self, env: str, spec: dict):
        nakadi = self._get_nakadi(env)
        et = EventType.from_spec(spec)

        try:
            current = nakadi.get_event_type(et.name)
            if current:
                logging.debug("Found existing %s", et)
                diff = DeepDiff(current, et, ignore_order=True, report_repetition=True)
                if diff:
                    self._maybe_print_diff(et, diff)
                    self._maybe_print_payload(et)
                    self._update_event_type(nakadi, et)

                else:
                    logging.info(f"{UP_TO_DATE_COLOR}✔ Up to date:{Fore.RESET} %s", et)

            else:
                logging.debug("Not found existing %s", et)
                self._maybe_print_payload(et)
                self._create_event_type(nakadi, et)

        except NakadiError as err:
            raise ProcessingError(f"Can not process {et}: {err}") from err

    def apply_sql_query(self, env: str, spec: dict):
        nakadi = self._get_nakadi(env)
        nakadi_sql = self._get_nakadi_sql(env)
        query = SqlQuery.from_spec(spec)

        try:
            current_et = nakadi.get_event_type(query.name)
            current = nakadi_sql.get_sql_query(current_et) if current_et else None
            if current:
                logging.debug("Found existing %s", query)

                diff = DeepDiff(
                    current, query, ignore_order=True, report_repetition=True
                )
                if diff:
                    self._maybe_print_diff(query, diff)

                    if "values_changed" in diff:
                        logging.info(
                            f"{ERROR_COLOR}× Modifying output event type is forbidden:{Fore.RESET} %s",
                            query,
                        )

                    else:
                        self._maybe_print_payload(query)
                        self._update_sql_query(nakadi_sql, query)

                else:
                    logging.info(
                        f"{UP_TO_DATE_COLOR}✔ Up to date:{Fore.RESET} %s", query
                    )

            else:
                logging.debug("Not found existing %s", query)
                self._maybe_print_payload(query)
                self._create_sql_query(nakadi, nakadi_sql, query)

        except NakadiError as err:
            raise ProcessingError(f"Can not process {query}: {err}") from err

    def apply_subscription(self, env: str, spec: dict):
        nakadi = self._get_nakadi(env)
        sub = Subscription.from_spec(spec)

        try:
            current = nakadi.get_subscription(
                sub.event_types, sub.owning_application, sub.consumer_group
            )
            if current:
                logging.debug("Found existing %s", current)
                sub.id = current.id
                diff = DeepDiff(current, sub, ignore_order=True, report_repetition=True)
                if diff:
                    self._maybe_print_diff(sub, diff)
                    self._maybe_print_payload(sub)
                    self._update_subscription(nakadi, sub)

                else:
                    logging.info(f"{UP_TO_DATE_COLOR}✔ Up to date:{Fore.RESET} %s", sub)

            else:
                logging.debug("Not found existing subscriptions matching: %s", sub)
                self._maybe_print_payload(sub)
                self._create_subscription(nakadi, sub)

        except NakadiError as err:
            raise ProcessingError(f"Can not process {sub}: {err}") from err

    def _maybe_print_diff(self, entity: Entity, diff: DeepDiff):
        def convert_to_spec(x):
            return x.to_spec()

        if self.show_diff:
            safe_diff = json.loads(
                diff.to_json(
                    default_mapping={
                        ReadOnlyAuth: convert_to_spec,
                        ReadWriteAuth: convert_to_spec,
                        EventType: convert_to_spec,
                        SqlQuery: convert_to_spec,
                        Subscription: convert_to_spec,
                        Entity: convert_to_spec,
                    }
                )
            )

            logging.info(
                f"{MODIFY_COLOR}⦿ Found %d changes:{Fore.RESET} %s\n%s",
                len(diff),
                entity,
                pretty_yaml(safe_diff, indentation=OUTPUT_INDENTATION),
            )

    def _maybe_print_payload(self, entity: Entity):
        if self.show_payload:
            if isinstance(entity, EventType):
                payload = event_type_to_payload(entity)
            elif isinstance(entity, Subscription):
                payload = subscription_to_payload(entity)
            elif isinstance(entity, SqlQuery):
                payload = sql_query_to_payload(entity)
            else:
                logging.warning("Failed to print payload for %s", entity)
                return

            logging.info(
                f"{MODIFY_COLOR}⦿ Nakadi payload:{Fore.RESET} %s\n%s",
                entity,
                pretty_json(payload, indentation=OUTPUT_INDENTATION),
            )

    def _update_event_type(self, nakadi: Nakadi, et: EventType):
        if self.execute:
            nakadi.update_event_type(et)
            logging.info(f"{MODIFY_COLOR}⦿ Updated:{Fore.RESET} %s", et)
        else:
            logging.info(f"{MODIFY_COLOR}⦿ Will update:{Fore.RESET} %s", et)

    def _create_event_type(self, nakadi: Nakadi, et: EventType):
        if self.execute:
            nakadi.create_event_type(et)
            logging.info(f"{MODIFY_COLOR}⦿ Created:{Fore.RESET} %s", et)
        else:
            logging.info(f"{MODIFY_COLOR}⦿ Will create:{Fore.RESET} %s", et)

    def _update_subscription(self, nakadi: Nakadi, sub: Subscription):
        if self.execute:
            nakadi.update_subscription(sub)
            logging.info(f"{MODIFY_COLOR}⦿ Updated:{Fore.RESET} %s", sub)
        else:
            logging.info(f"{MODIFY_COLOR}⦿ Will update:{Fore.RESET} %s", sub)

    def _create_subscription(self, nakadi: Nakadi, sub: Subscription):
        if self.execute:
            nakadi.create_subscription(sub)
            logging.info(f"{MODIFY_COLOR}⦿ Created:{Fore.RESET} %s", sub)
        else:
            logging.info(f"{MODIFY_COLOR}⦿ Will create:{Fore.RESET} %s", sub)

    def _update_sql_query(self, nakadi_sql: NakadiSql, query: SqlQuery):
        if self.execute:
            nakadi_sql.update_sql_query_auth(query)
            logging.info(f"{MODIFY_COLOR}⦿ Updated:{Fore.RESET} %s", query)
        else:
            logging.info(f"{MODIFY_COLOR}⦿ Will update:{Fore.RESET} %s", query)

    def _create_sql_query(self, nakadi: Nakadi, nakadi_sql: NakadiSql, query: SqlQuery):
        if self.execute:
            nakadi_sql.create_sql_query(query)
            logging.info(f"{MODIFY_COLOR}⦿ Created:{Fore.RESET} %s", query)
        else:
            logging.info(f"{MODIFY_COLOR}⦿ Will create:{Fore.RESET} %s", query)

    def _get_nakadi(self, env: str) -> Nakadi:
        if env not in self.config.environments:
            raise ProcessingError(f"Unknown environment: {env}")
        return Nakadi(self.config.environments[env].nakadi_url, self.token)

    def _get_nakadi_sql(self, env: str) -> NakadiSql:
        if env not in self.config.environments:
            raise ProcessingError(f"Unknown environment: {env}")
        if not self.config.environments[env].nakadi_sql_url:
            raise ProcessingError("Nakadi SQL endpoint is not configured")
        return NakadiSql(self.config.environments[env].nakadi_sql_url, self.token)


class ProcessingError(Exception):
    def __init__(self, msg: str):
        self.msg = msg

    def __str__(self):
        return self.msg



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import yaml


@dataclass
class EnvironmentConfig:
    nakadi_url: str
    nakadi_sql_url: Optional[str]

    @staticmethod
    def load(name: str, content: dict[str, str]) -> EnvironmentConfig:
        if "nakadi_url" not in content:
            raise ConfigurationError(
                f"Nakadi url not found in configuration for environment: {name}"
            )

        return EnvironmentConfig(
            nakadi_url=content["nakadi_url"],
            nakadi_sql_url=content.get("nakadi_sql_url", None),
        )


@dataclass
class AppConfig:
    environments: dict[str, EnvironmentConfig]

    @staticmethod
    def load(content: dict) -> AppConfig:
        if "environments" not in content:
            raise ConfigurationError(f"Environments section not found in configuration")

        return AppConfig(
            {
                name: EnvironmentConfig.load(name, val)
                for name, val in content["environments"].items()
            }
        )


def load_config() -> AppConfig:
    config_locations = [Path.cwd(), Path.home()]

    for file_name in [str(path / ".clin") for path in config_locations]:
        if os.path.isfile(file_name):
            try:
                with open(file_name) as f:
                    content = yaml.safe_load(f)
            except Exception as ex:
                raise ConfigurationError(
                    f"Failed to parse configuration file: {file_name}: {str(ex)}"
                )

            return AppConfig.load(content)

    raise ConfigurationError(
        f"No valid configuration file found. Inspected locations: "
        f"{[str(path) for path in config_locations]}"
    )


class ConfigurationError(Exception):
    def __init__(self, message: str):
        self.message = message
